@@ -25,6 +25,8 @@ def _docsubst(f):
     f.__doc__ = f.__doc__.format(**_doc_snippets)
     return f
 
+_seq_sn = "If an (S, N) array, the first dimension will be used for frames in an animation."
+_seq_snm = "If an (S, N, M) array, the first dimension will be used for frames in an animation."
 
 _doc_snippets = {}
 _doc_snippets[
@@ -34,15 +36,15 @@ _doc_snippets[
 _doc_snippets[
     "size"] = "float representing the size of the glyph in percentage of the viewport, where 100 is the full size of the viewport"
 _doc_snippets["marker"] = "name of the marker, options are: 'arrow', 'box', 'diamond', 'sphere'"
-_doc_snippets["x"] = "numpy array of shape (N,) or (S, N) with x positions (can be a sequence)"
+_doc_snippets["x"] = "numpy array of shape (N,) or (S, N) with x positions. {}".format(_seq_sn)
 _doc_snippets["y"] = "idem for y"
 _doc_snippets["z"] = "idem for z"
-_doc_snippets["u_dir"] = "numpy array of shape (N,) or (S, N) indicating the x component of a vector (can be a sequence)"
+_doc_snippets["u_dir"] = "numpy array of shape (N,) or (S, N) indicating the x component of a vector. {}".format(_seq_sn)
 _doc_snippets["v_dir"] = "idem for y"
 _doc_snippets["w_dir"] = "idem for z"
-_doc_snippets["u"] = "numpy array of shape (N,) or (S, N) indicating the u (x) coordinate for the texture (can be a sequence)"
-_doc_snippets["v"] = "numpy array of shape (N,) or (S, N) indicating the v (y) coordinate for the texture (can be a sequence)"
-_doc_snippets["x2d"] = "numpy array of shape (N,M) or (S, N, M) with x positions (can be a sequence)"
+_doc_snippets["u"] = "numpy array of shape (N,) or (S, N) indicating the u (x) coordinate for the texture. {}".format(_seq_sn)
+_doc_snippets["v"] = "numpy array of shape (N,) or (S, N) indicating the v (y) coordinate for the texture. {}".format(_seq_sn)
+_doc_snippets["x2d"] = "numpy array of shape (N,M) or (S, N, M) with x positions. {}".format(_seq_snm)
 _doc_snippets["y2d"] = "idem for y"
 _doc_snippets["z2d"] = "idem for z"
 _doc_snippets["texture"] = "PIL.Image object or ipywebrtc.MediaStream (can be a seqence)"
@@ -327,7 +329,7 @@ def plot_mesh(x, y, z, color=default_color, wireframe=True, surface=True, wrapx=
     def index_from2d(i, j):
         xi = (i % nx)
         yi = (j % ny)
-        return nx * xi + yi
+        return ny * xi + yi
         """
         ^ ydir
         |
@@ -341,7 +343,7 @@ def plot_mesh(x, y, z, color=default_color, wireframe=True, surface=True, wrapx=
             p1 = index_from2d(i + 1, j)
             p2 = index_from2d(i, j + 1)
             p3 = index_from2d(i + 1, j + 1)
-            triangle_index = (i * mx) + j
+            triangle_index = (i * my) + j
             triangles[triangle_index * 2 + 0, :] = [p0, p1, p3]
             triangles[triangle_index * 2 + 1, :] = [p0, p3, p2]
             lines[triangle_index * 4 + 0, :] = [p0, p1]
@@ -368,16 +370,19 @@ def plot(x, y, z, color=default_color, **kwargs):
     """
     fig = gcf()
     _grow_limits(x, y, z)
-    scatter = ipv.Scatter(x=x, y=y, z=z, color=color, color_selected=None,
-        size_selected=1, size=1,
-                             connected=True, visible_markers=False, visible_lines=True, **kwargs)
-    fig.scatters = fig.scatters + [scatter]
-    return scatter
+    defaults = dict(visible_lines=True, color_selected=None, size_selected=1,
+                    size=1, connected=True, visible_markers=False)
+    kwargs = dict(defaults, **kwargs)
+    s = ipv.Scatter(x=x, y=y, z=z, color=color, **kwargs)
+    fig.scatters = fig.scatters + [s]
+    return s
 
 
 @_docsubst
-def scatter(x, y, z, color=default_color, size=default_size, size_selected=default_size_selected,
-            color_selected=default_color_selected, marker="diamond", selection=None, **kwargs):
+def scatter(x, y, z, color=default_color, size=default_size,
+            size_selected=default_size_selected,
+            color_selected=default_color_selected, marker="diamond",
+            selection=None, **kwargs):
     """Plots many markers/symbols in 3d
 
     :param x: {x}
@@ -394,13 +399,15 @@ def scatter(x, y, z, color=default_color, size=default_size, size_selected=defau
     """
     fig = gcf()
     _grow_limits(x, y, z)
-    scatter = ipv.Scatter(x=x, y=y, z=z, color=color, size=size, color_selected=color_selected,
-                             size_selected=size_selected, geo=marker, selection=selection, **kwargs)
-    fig.scatters = fig.scatters + [scatter]
-    return scatter
+    s = ipv.Scatter(x=x, y=y, z=z, color=color, size=size,
+                    color_selected=color_selected, size_selected=size_selected,
+                    geo=marker, selection=selection, **kwargs)
+    fig.scatters = fig.scatters + [s]
+    return s
 
 @_docsubst
-def quiver(x, y, z, u, v, w, size=default_size * 10, size_selected=default_size_selected * 10, color=default_color,
+def quiver(x, y, z, u, v, w, size=default_size * 10,
+           size_selected=default_size_selected * 10, color=default_color,
            color_selected=default_color_selected, marker="arrow", **kwargs):
     """Create a quiver plot, which is like a scatter plot but with arrows pointing in the direction given by u, v and w
 
@@ -420,11 +427,13 @@ def quiver(x, y, z, u, v, w, size=default_size * 10, size_selected=default_size_
     """
     fig = gcf()
     _grow_limits(x, y, z)
-    scatter = ipv.Scatter(x=x, y=y, z=z, vx=u, vy=v, vz=w,
-                             color=color, size=size, color_selected=color_selected, size_selected=size_selected,
-                             geo=marker, **kwargs)
-    fig.scatters = fig.scatters + [scatter]
-    return scatter
+    if 'vx' in kwargs or 'vy' in kwargs or 'vz' in kwargs:
+        raise KeyError('Please use u, v, w instead of vx, vy, vz')
+    s = ipv.Scatter(x=x, y=y, z=z, vx=u, vy=v, vz=w, color=color, size=size,
+                    color_selected=color_selected, size_selected=size_selected,
+                    geo=marker, **kwargs)
+    fig.scatters = fig.scatters + [s]
+    return s
 
 
 def show(extra_widgets=[]):
@@ -447,7 +456,7 @@ def animation_control(object, sequence_length=None, add=True, interval=200):
     :param object: :any:`Scatter` or :any:`Mesh` object (having an sequence_index property), or a list of these to control multiple.
     :param sequence_length: If sequence_length is None we try try our best to figure out, in case we do it badly,
             you can tell us what it should be. Should be equal to the S in the shape of the numpy arrays as for instance documented
-            in :any:`scatter` or :any:`plot_mesh`. 
+            in :any:`scatter` or :any:`plot_mesh`.
     :param add: if True, add the widgets to the container, else return a HBox with the slider and play button. Useful when you
             want to customise the layout of the widgets yourself.
     :param interval: interval in msec between each frame
@@ -546,7 +555,6 @@ def plot_isosurface(data, level=None, color=default_color, wireframe=True, surfa
         values = measure.marching_cubes_lewiner(data, level)
     else:
         values = measure.marching_cubes(data, level)
-    values = measure.marching_cubes_lewiner(data, level)#, spacing=(0.1, 0.1, 0.1))
     verts, triangles = values[:2] # version 0.13 returns 4 values, normals, values
     # in the future we may want to support normals and the values (with colormap)
     # and require skimage >= 0.13
@@ -554,7 +562,8 @@ def plot_isosurface(data, level=None, color=default_color, wireframe=True, surfa
     mesh = plot_trisurf(x, y, z, triangles=triangles, color=color)
     if controls:
         vmin, vmax = np.percentile(data, 1),  np.percentile(data, 99)
-        level_slider = ipywidgets.FloatSlider(value=level, min=vmin, max=vmax, icon='eye')
+        step = (vmax - vmin)/250
+        level_slider = ipywidgets.FloatSlider(value=level, min=vmin, max=vmax, step=step, icon='eye')
         recompute_button = ipywidgets.Button(description='update')
         controls = ipywidgets.HBox(children=[level_slider, recompute_button])
         current.container.children += (controls,)
@@ -565,7 +574,6 @@ def plot_isosurface(data, level=None, color=default_color, wireframe=True, surfa
                 values = measure.marching_cubes_lewiner(data, level)
             else:
                 values = measure.marching_cubes(data, level)
-            values = measure.marching_cubes_lewiner(data, level)#, spacing=(0.1, 0.1, 0.1))
             verts, triangles = values[:2] # version 0.13 returns 4 values, normals, values
             # in the future we may want to support normals and the values (with colormap)
             # and require skimage >= 0.13
@@ -736,49 +744,60 @@ def movie(f="movie.mp4", function=_change_y_angle, fps=30, frames=30, endpoint=F
     return tempdir
 
 
-def _screenshot_data(timeout_seconds=10, output_widget=None, format="png", width=None, height=None, fig=None):
+def _screenshot_data(timeout_seconds=10, output_widget=None, format="png", width=None, height=None, fig=None, headless=False):
     if fig is None:
         fig = gcf()
     else:
         assert isinstance(fig, ipv.Figure)
-    if output_widget is None:
-        output_widget = ipywidgets.Output()
-        display(output_widget)
-    # use lists to avoid globals
-    done = [False]
-    data = [None]
+    if headless:
+        from . import headless
+        import tempfile
+        tempdir = tempfile.mkdtemp()
+        tempdir = '/var/folders/vn/_rmzj8jd0215_g9yfrn8pmgm0000gn/T/tmpxv07_m6l/'
+        tempfile = os.path.join(tempdir, 'headless.html')
+        #tempfile = os.path.abspath(tempfile)
+        save(tempfile, offline=True, scripts_path=tempdir, devmode=True)
+        print(tempfile)
+        data = headless._screenshot_data("file://" + tempfile)
+    else:
+        if output_widget is None:
+            output_widget = ipywidgets.Output()
+            display(output_widget)
+        # use lists to avoid globals
+        done = [False]
+        data = [None]
 
-    def screenshot_handler(image_data):
-        with output_widget:
-            # print("data")
-            # print(data)
-            done[0] = True
-            data[0] = image_data
-
-    fig.on_screenshot(screenshot_handler)
-    try:
-        fig.screenshot(width=width, height=height,mime_type="image/"+format)
-        t0 = time.time()
-        timeout = False
-        ipython = IPython.get_ipython()
-        while (not done[0]) and not timeout:
-            ipython.kernel.do_one_iteration()
+        def screenshot_handler(image_data):
             with output_widget:
-                time.sleep(0.05)
-                timeout = (time.time() - t0) > timeout_seconds
-        with output_widget:
-            if timeout and not done[0]:
-                raise ValueError("timed out, no image data returned")
-    finally:
-        with output_widget:
-            fig.on_screenshot(screenshot_handler, remove=True)
-    data = data[0]
+                # print("data")
+                # print(data)
+                done[0] = True
+                data[0] = image_data
+
+        fig.on_screenshot(screenshot_handler)
+        try:
+            fig.screenshot(width=width, height=height,mime_type="image/"+format)
+            t0 = time.time()
+            timeout = False
+            ipython = IPython.get_ipython()
+            while (not done[0]) and not timeout:
+                ipython.kernel.do_one_iteration()
+                with output_widget:
+                    time.sleep(0.05)
+                    timeout = (time.time() - t0) > timeout_seconds
+            with output_widget:
+                if timeout and not done[0]:
+                    raise ValueError("timed out, no image data returned")
+        finally:
+            with output_widget:
+                fig.on_screenshot(screenshot_handler, remove=True)
+        data = data[0]
     data = data[data.find(",") + 1:]
     return base64.b64decode(data)
 
-def screenshot(width=None, height=None, format="png", fig=None, timeout_seconds=10, output_widget=None):
+def screenshot(width=None, height=None, format="png", fig=None, timeout_seconds=10, output_widget=None, headless=False):
     """Save the figure to a PIL.Image object.
-    
+
     :param int width: the width of the image in pixels
     :param int height: the height of the image in pixels
     :param format: format of output data (png, jpeg or svg)
@@ -787,33 +806,33 @@ def screenshot(width=None, height=None, format="png", fig=None, timeout_seconds=
     :type timeout_seconds: int
     :param timeout_seconds: maximum time to wait for image data to return
     :type output_widget: ipywidgets.Output
-    :param output_widget: a widget to use as a context manager for capturing the data  
+    :param output_widget: a widget to use as a context manager for capturing the data
     :return: PIL.Image
 
     """
     assert format in ['png','jpeg','svg'], "image format must be png, jpeg or svg"
-    data = _screenshot_data(timeout_seconds=timeout_seconds, output_widget=output_widget, 
-    format=format, width=width, height=height, fig=fig)
+    data = _screenshot_data(timeout_seconds=timeout_seconds, output_widget=output_widget,
+    format=format, width=width, height=height, fig=fig, headless=headless)
     f = StringIO(data)
     return PIL.Image.open(f)
 
-def savefig(filename, width=None, height=None, fig=None, timeout_seconds=10, output_widget=None):
+def savefig(filename, width=None, height=None, fig=None, timeout_seconds=10, output_widget=None, headless=False):
     """Save the figure to an image file.
-    
+
     :param str filename: must have extension .png, .jpeg or .svg
     :param int width: the width of the image in pixels
     :param int height: the height of the image in pixels
     :type fig: ipyvolume.widgets.Figure or None
-    :param fig: if None use the current figure    
+    :param fig: if None use the current figure
     :param float timeout_seconds: maximum time to wait for image data to return
-    :param ipywidgets.Output output_widget: a widget to use as a context manager for capturing the data  
+    :param ipywidgets.Output output_widget: a widget to use as a context manager for capturing the data
     """
     __, ext = os.path.splitext(filename)
     format = ext[1:]
     assert format in ['png','jpeg','svg'], "image format must be png, jpeg or svg"
     with open(filename, "wb") as f:
-        f.write(_screenshot_data(timeout_seconds=timeout_seconds, output_widget=output_widget, 
-        format=format, width=width, height=height, fig=fig))
+        f.write(_screenshot_data(timeout_seconds=timeout_seconds, output_widget=output_widget,
+        format=format, width=width, height=height, fig=fig, headless=headless))
 
 
 def xlabel(label):
